package com.f23ko.gson;

import java.lang.reflect.Type;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.google.gson.JsonDeserializationContext;
import com.google.gson.JsonDeserializer;
import com.google.gson.JsonElement;
import com.google.gson.JsonParseException;
import com.google.gson.JsonPrimitive;
import com.google.gson.JsonSerializationContext;
import com.google.gson.JsonSerializer;
import com.google.gson.JsonSyntaxException;

/**
 * This type adapter supports three subclasses of date: Date, Timestamp, and
 * java.sql.Date.
 *
 * @author Inderjeet Singh
 * @author Joel Leitch
 */
public final class F23DateTypeAdapter implements JsonSerializer<Date>, JsonDeserializer<Date> 
{
  private final DateFormat iso8601Format;
  private final boolean ftime;
  private final Pattern pattern;
  private final String timeTag;
  public F23DateTypeAdapter() {
	this("FDate");
  }
  public F23DateTypeAdapter(String timeTag) {
	this.ftime=timeTag!=null;
	this.timeTag=timeTag;
	this.pattern=ftime?Pattern.compile("^"+timeTag+"\\(([0-9]+)\\)$"):null;
    this.iso8601Format = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'", Locale.US);
    this.iso8601Format.setTimeZone(TimeZone.getTimeZone("UTC"));
  }

  // These methods need to be synchronized since JDK DateFormat classes are not thread-safe
  // See issue 162
  public JsonElement serialize(Date src, Type typeOfSrc, JsonSerializationContext context) {
	if(ftime)
	{
		return new JsonPrimitive(timeTag+"("+src.getTime()+")");
	}
	else
	{
	    synchronized (iso8601Format) {
	      String dateFormatAsString = iso8601Format.format(src);
	      return new JsonPrimitive(dateFormatAsString);
	    }
	}
  }

  public Date deserialize(JsonElement json, Type typeOfT, JsonDeserializationContext context) throws JsonParseException 
  {
    if (!(json instanceof JsonPrimitive)) {
      throw new JsonParseException("The date should be a string value");
    }
    Date date = deserializeToDate(json);
    if (typeOfT == Date.class) {
      return date;
    } else if (typeOfT == Timestamp.class) {
      return new Timestamp(date.getTime());
    } else if (typeOfT == java.sql.Date.class) {
      return new java.sql.Date(date.getTime());
    } else {
      throw new IllegalArgumentException(getClass() + " cannot deserialize to " + typeOfT);
    }
  }

  private Date deserializeToDate(JsonElement json) {
	  String date=json.getAsString();
	  Matcher matcher=  pattern.matcher(date);
	  if(matcher.find())
	  {
		Date d=new Date();
		d.setTime(Long.parseLong(matcher.group(0)));
		return d;
	  }
	  synchronized (iso8601Format) {
	     try {
	        return iso8601Format.parse(date);
	     } catch (ParseException e) {
	        throw new JsonSyntaxException(json.getAsString(), e);
	     }
	  }
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append(F23DateTypeAdapter.class.getSimpleName());
    sb.append('(').append(iso8601Format.getClass().getSimpleName()).append(')');
    return sb.toString();
  }
}
