package com.f23ko.gson;

/*
 * Copyright (C) 2011 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */



import com.google.gson.FieldNamingPolicy;
import com.google.gson.FieldNamingStrategy;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.InstanceCreator;
import com.google.gson.JsonSyntaxException;
import com.google.gson.TypeAdapter;
import com.google.gson.TypeAdapterFactory;
import com.google.gson.annotations.SerializedName;
import com.google.gson.internal.$Gson$Types;
import com.google.gson.internal.ConstructorConstructor;
import com.google.gson.internal.Excluder;
import com.google.gson.internal.ObjectConstructor;
import com.google.gson.internal.Primitives;
import com.google.gson.reflect.TypeToken;
import com.google.gson.stream.JsonReader;
import com.google.gson.stream.JsonToken;
import com.google.gson.stream.JsonWriter;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.Type;
import java.net.URL;
import java.sql.Timestamp;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 * Type adapter that reflects over the fields and methods of a class.
 */
public final class FDataTypeAdapterFactory implements TypeAdapterFactory {
  private final ConstructorConstructor constructorConstructor;
  private final FieldNamingStrategy fieldNamingPolicy;
  private final Excluder excluder;
  private final String  objKey;
  private final String  packagePath;
  private final Map<String,Class<?>>  dataClassMap;
  //private final Map<String,ObjectConstructor<?>>  dataConstructorMap;
  private final Map<Class<?>,String>  dataNameMap;
  private final Class<?> tag;

  /**
   * 
   * @param packagePath 你的資料物件的資料夾例如com.f23ko.data，要傳輸的物件請實做 com.f23ko.gson.IData 介面
   */
  public FDataTypeAdapterFactory(String packagePath)
  {
	this("objKey",IData.class,packagePath);
  }
  
  /**
   * 初始化套件轉換工廠
   * @param objName 識別物件的名稱，請確定這個不會跟你自己的物件撞名
   * @param packagePath 讀取物件的套件位置，你應該把允許自動識別的物件都放在這個套件當中，並繼承IData的介面，例如 com.f23ko.json.data
   */
  public FDataTypeAdapterFactory(String objKey,Class<?> tag,String packagePath)
  {
	this.tag=tag;
	this.constructorConstructor = new ConstructorConstructor(Collections.<Type, InstanceCreator<?>>emptyMap());
    this.fieldNamingPolicy = FieldNamingPolicy.IDENTITY;
    this.excluder = Excluder.DEFAULT;
    this.objKey=objKey;
    this.packagePath=packagePath;
    try
	{
		dataClassMap=Collections.unmodifiableMap(getClassInPackage(packagePath));
	}
	catch (ClassNotFoundException e)
	{
		throw new RuntimeException(e);
	}
    Map<Class<?>,String> dataNameMap=new HashMap<>();
    for(Map.Entry<String, Class<?>> me: dataClassMap.entrySet())
    {
    	dataNameMap.put(me.getValue(), me.getKey());
    }
    this.dataNameMap=Collections.unmodifiableMap(dataNameMap);
  }

  public boolean excludeField(Field f, boolean serialize)
  {
    return !excluder.excludeClass(f.getType(), serialize) && !excluder.excludeField(f, serialize);
  }

  private String getFieldName(Field f)
  {
    SerializedName serializedName = f.getAnnotation(SerializedName.class);
    return serializedName == null ? fieldNamingPolicy.translateName(f) : serializedName.value();
  }

  public <T> TypeAdapter<T> create(Gson gson, final TypeToken<T> type) 
  {
	//是IData才轉換
	Class<? super T> raw = type.getRawType();
    if (!tag.isAssignableFrom(raw)) 
    {
      return null; 
    }

	Map<String,ClassHelper> dataClassHelpers=new HashMap<>();
    for(Map.Entry<String, Class<?>> me: dataClassMap.entrySet())
    {
    	TypeToken<?> t= TypeToken.get(me.getValue());
    	Map<String, BoundField> bf=getBoundFields(gson, t, me.getValue());
    	dataClassHelpers.put(me.getKey(),new ClassHelper(bf,constructorConstructor.get(t)));
    }
    dataClassHelpers=Collections.unmodifiableMap(dataClassHelpers);    
    return new Adapter<T>(objKey,dataClassHelpers,dataNameMap);
  }

  private class ClassHelper
  {
	Map<String, BoundField> boundFields;
	ObjectConstructor<?> constructor;
	public ClassHelper(Map<String, BoundField> boundFields,
			ObjectConstructor<?> constructor)
	{
		this.boundFields = boundFields;
		this.constructor = constructor;
	}
  }
  private FDataTypeAdapterFactory.BoundField createBoundField( final Gson context, final Field field, final String name, final TypeToken<?> fieldType, boolean serialize, boolean deserialize)
  {
    final boolean isPrimitive = Primitives.isPrimitive(fieldType.getRawType());

    // special casing primitives here saves ~5% on Android...
    return new FDataTypeAdapterFactory.BoundField(name, serialize, deserialize) {
      final TypeAdapter<?> typeAdapter = context.getAdapter(fieldType);
      @SuppressWarnings({"unchecked", "rawtypes"}) // the type adapter and field type always agree
      @Override void write(JsonWriter writer, Object value) throws IOException, IllegalAccessException 
      {
        Object fieldValue = field.get(value);
        TypeAdapter t =new FTypeAdapterRuntimeTypeWrapper(context, this.typeAdapter, fieldType.getType());
        t.write(writer, fieldValue);
      }
      @Override void read(JsonReader reader, Object value)
          throws IOException, IllegalAccessException {
        Object fieldValue = typeAdapter.read(reader);
        if (fieldValue != null || !isPrimitive) {
          field.set(value, fieldValue);
        }
      }
    };
  }

  private Map<String, BoundField> getBoundFields(Gson context, TypeToken<?> type, Class<?> raw) {
    Map<String, BoundField> result = new LinkedHashMap<String, BoundField>();
    if (raw.isInterface()) {
      return result;
    }

    Type declaredType = type.getType();
    while (raw != Object.class) 
    {
      Field[] fields = raw.getDeclaredFields();
      for (Field field : fields) 
      {
        boolean serialize = excludeField(field, true);
        boolean deserialize = excludeField(field, false);
        if (!serialize && !deserialize) 
        {
          continue;
        }
        field.setAccessible(true);
        Type fieldType = $Gson$Types.resolve(type.getType(), raw, field.getGenericType());
        BoundField boundField = createBoundField(context, field, getFieldName(field),TypeToken.get(fieldType), serialize, deserialize);
        BoundField previous = result.put(boundField.name, boundField);
        if (previous != null)
        {
          throw new IllegalArgumentException(declaredType + " declares multiple JSON fields named " + previous.name);
        }
      }
      type = TypeToken.get($Gson$Types.resolve(type.getType(), raw, raw.getGenericSuperclass()));
      raw = type.getRawType();
    }
    return result;
  }

  static abstract class BoundField {
    final String name;
    final boolean serialized;
    final boolean deserialized;

    protected BoundField(String name, boolean serialized, boolean deserialized) {
      this.name = name;
      this.serialized = serialized;
      this.deserialized = deserialized;
    }

    abstract void write(JsonWriter writer, Object value) throws IOException, IllegalAccessException;
    abstract void read(JsonReader reader, Object value) throws IOException, IllegalAccessException;
  }
/**
 * 取得該 package 下的所有 Class 檔案
 * @param pckgname 套件的位置
 * @return class列表
 * @throws ClassNotFoundException 
 */
private static Map<String,Class<?>> getClassInPackage(String pckgname) throws ClassNotFoundException
{
	Map<String,Class<?>> classes = new HashMap<>();
    // Get a File object for the package
    File directory = null;
    try {
        ClassLoader cld = Thread.currentThread().getContextClassLoader();
        if (cld == null) {
            throw new ClassNotFoundException("Can't get class loader.");
        }
        String path = pckgname.replace('.', '/');
        URL resource = cld.getResource(path);
        if (resource == null) {
            throw new ClassNotFoundException("No resource for " + path);
        }
        directory = new File(resource.getFile());
    }
    catch (NullPointerException x) {
        throw new ClassNotFoundException(pckgname + " (" + directory + ") does not appear to be a valid package");
    }
    if (directory.exists()) {
        // Get the list of the files contained in the package
        String[] files = directory.list();
        for (int i = 0; i < files.length; i++) {
            // we are only interested in .class files
            if (files[i].endsWith(".class")) {
                // removes the .class extension
            	String name=files[i].substring(0, files[i].length() - 6);
                classes.put(name,Class.forName(pckgname + '.' + name));
            }
        }
    }
    else {
        throw new ClassNotFoundException(pckgname + " does not appear to be a valid package");
    }
    return classes;
}
	
  public static final class Adapter<T> extends TypeAdapter<T> 
  {
    private final String objKey;
    private final Map<String,ClassHelper> classHelpers;
    private final Map<Class<?>,String>  dataNameMap;

    private Adapter(String objKey,Map<String,ClassHelper> classHelpers,Map<Class<?>,String>  dataNameMap) {
      this.dataNameMap=dataNameMap;
      this.objKey=objKey;
      this.classHelpers=classHelpers;
    }

    @Override public T read(JsonReader in) throws IOException {
      if (in.peek() == JsonToken.NULL) {
        in.nextNull();
        return null;
      }

      T instance =null;
      try {
        in.beginObject();
        if(!in.hasNext()) throw new RuntimeException("JSON object is empty boject");
        String objKey = in.nextName();
        
        if(!this.objKey.equals(objKey)) throw new JsonSyntaxException("JSON object does not has ObjKey in the first");
        
        ObjectConstructor<T> constructor=null;
        Map<String, BoundField> boundFields=null;
        try{
 	      String objName=in.nextString();
 	      ClassHelper classHelper=classHelpers.get(objName);
	      constructor=(ObjectConstructor<T>) classHelper.constructor;
	      boundFields=classHelper.boundFields;
	      instance=constructor.construct();

        } catch (RuntimeException e) {
        	throw new JsonSyntaxException("Can not Creat instance. makesure your class is in packagePath and implement tag interfce(defult IData) ",e);
        }
        

        while (in.hasNext()) {
          String name = in.nextName();      
          BoundField field = boundFields.get(name);
          if (field == null || !field.deserialized) {
            in.skipValue();
          } else {
            field.read(in, instance);
          }
        }
      } catch (IllegalStateException e) {
        throw new JsonSyntaxException(e);
      } catch (IllegalAccessException e) {
        throw new AssertionError(e);
      }
      in.endObject();
      return instance;
    }

    @Override public void write(JsonWriter out, T value) throws IOException {
      if (value == null) {
        out.nullValue();
        return;
      }

      //寫入物件的型別標籤
      out.beginObject();
      out.name(objKey);
      String name=dataNameMap.get(value.getClass());
      if(name==null) throw new JsonSyntaxException("Class("+value.getClass().getName()+") do not in ALLOW List");
      out.value(name);
      
      Map<String, BoundField> boundFields=classHelpers.get(name).boundFields;
     
      try {
        for (BoundField boundField : boundFields.values()) {
          if (boundField.serialized) {
            out.name(boundField.name);
            boundField.write(out, value);
          }
        }
      } catch (IllegalAccessException e) {
        throw new AssertionError();
      }
      out.endObject();
    }
  }
  public static void main(String[] args) throws ParseException
  {
	  //EXAMPLE
	  F23DateTypeAdapter fDate=new F23DateTypeAdapter();
	  Gson gson=new GsonBuilder()
		.registerTypeAdapterFactory(new FDataTypeAdapterFactory("com.f23ko.test"))
		.registerTypeAdapter(Date.class,fDate)
		.registerTypeAdapter(Timestamp.class,fDate)
		.registerTypeAdapter(java.sql.Date.class,fDate)
		.create();
		  
	  
  }
}
